import abc

from basics.base import Base


class UnableToGenerateValidResponse(Exception):
    pass


class ConversationEngineBase(Base, metaclass=abc.ABCMeta):

    def __init__(self, model_evaluator, debug=False, name=None):
        super().__init__(pybase_logger_name=name)

        self._model_evaluator = model_evaluator

        self._debug = debug

        self._conversation_context = {}

    def reset_state(self):
        self._conversation_context = {}
        self._model_evaluator.reset_state()

    def execute_command(self, command, user_name=None):
        """

        :param command:
        :param user_name: Optional, user name of the user who input the command

        :return: <system message> = None, when the conversation engine did not process any command
        """
        return None

    @abc.abstractmethod
    def respond(self, inputs, conversation_start=False):
        """

        :param inputs: Dict with one or more different toes of inputs
        :param conversation_start: Boolean

        :return: <response to input>, <score(s)>, <other outputs or None>
        """
        self._log.error("Please implement this method in a child class")


class BasicConversationEngine(ConversationEngineBase):

    def __init__(self,
                 io_processor,
                 model_evaluator,
                 select_token_func,
                 sequence_end_token=None,
                 is_sequence_end_func=None,
                 max_response_length=-1,
                 **kwargs):
        super().__init__(model_evaluator, **kwargs)

        self._select_token_func = select_token_func

        if sequence_end_token is None and is_sequence_end_func is None:
            raise Exception("Define sequence_end_token or is_sequence_end_func")

        if sequence_end_token is not None and is_sequence_end_func is not None:
            raise Exception("Either define sequence_end_token or is_sequence_end_func")

        if sequence_end_token is not None:
            self._log.debug("Creating is_sequence_end_func using sequence_end_token")
            is_sequence_end_func = lambda response, scores: response[-1] == sequence_end_token

        self._io_processor = io_processor

        self.sequence_end_token = sequence_end_token
        self._is_sequence_end_func = is_sequence_end_func

        self._max_response_length = max_response_length

    def respond(self, inputs, conversation_start=False):
        processed_inputs = self._io_processor.process_inputs(inputs, conversation_start)

        self._model_evaluator.update_context(processed_inputs, self._conversation_context, conversation_start)

        return self._create_response()

    def _create_response(self, **kwargs):
        """
        Called after model context updated

        :return:
        :rtype:
        """

        self._will_create_response()

        prediction_context = {}
        prev_token = None
        response = []
        scores = []
        while True:
            prediction_data = self._model_evaluator.predict_next_token(prev_token,
                                                                       prediction_context,
                                                                       self._conversation_context)
            # Obtain most likely word token and its score
            score, token = self._select_token_func(prediction_data)

            # Add new token and score
            response += [self._unwrap(token)]
            scores += [self._unwrap(score)]

            if self._is_sequence_end(response, scores):
                break

            if len(response) >= self._max_response_length > 0:
                break

            prev_token = token

        response, scores = self._process_response(response, scores)

        return response, scores, None # Other outpus

    def _will_create_response(self):
        pass

    def _is_sequence_end(self, response, scores):
        return self._is_sequence_end_func(response=response, scores=scores)

    def _process_response(self, response, scores):
        return self._io_processor.process_response(response, scores=scores)

    def _unwrap(self, tensor):
        """
        Unwrap scalar tensor

        :param tensor:
        :return:
        """
        return tensor



import abc

from basics.base import Base


class IOProcessorBase(Base, metaclass=abc.ABCMeta):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def reset_state(self):
        pass

    @abc.abstractmethod
    def process_inputs(self, inputs, conversation_start):
        """

        :param inputs: Dict with one or more different types of inputs
        :param conversation_start: Bool, if the given inputs are the initial inputs of a new
                                   conversation

        :return:
        """
        self._log.error("Please implement this method in a child class")

    @abc.abstractmethod
    def process_response(self, response, scores=None, stop_sequence=None):
        """

        :param response:
        :param scores:
        :param stop_sequence:

        :return: processed response, scores
        """
        self._log.error("Please implement this method in a child class")


class SimpleIOProcessor(IOProcessorBase):

    def __init__(self,
                 encoding_func,
                 decoding_func,
                 preprocessing_func=None,
                 **kwargs):
        """

        :param encoding_func       : function that takes a list of sequences as inputs
                                     (not tokenized) and returns a tokenized and indexed list of inputs

        :param decoding_func     : function that takes a list of tokenized and indexed inputs
                                   and returns a detokenized and deindexed list of inputs

        :param preprocessing_func  : Optional, function that takes a list of sequences as inputs and
                                     returns a preprocessed list of inputs
        """
        super().__init__(**kwargs)

        self._encoding_func = encoding_func
        self._decoding_func = decoding_func

        self._preprocessing_func = preprocessing_func

    def process_inputs(self, inputs, conversation_start):
        """

        The given inputs are split in to sequence_inputs and other_inputs.
        The sequence_inputs are further processed, before both are returned.

        :param inputs: Dict with one or more different types of inputs
        :param conversation_start: Bool, if the given inputs are the initial inputs of a new
                                   conversation

        :return: processed sequence_inputs, other_inputs
        """
        sequence_inputs, other_inputs = self._get_input_chats(inputs, conversation_start)

        # Preprocessing is optional
        if callable(self._preprocessing_func):
            sequence_inputs = self._preprocessing_func(sequence_inputs)

        sequence_inputs = self._build_input_sequences(sequence_inputs, conversation_start)

        sequence_inputs = self.encode(sequence_inputs)

        sequence_inputs = self._postprocess(sequence_inputs)

        return sequence_inputs, other_inputs

    def process_response(self, response, scores=None, stop_sequence=None, **kwargs):
        """

        :param response: bot response
        :param scores: data structure (e.g. list) with one or more scores related to the response
        :param stop_sequence

        :return:
        """

        return self.decode(response), scores

    def encode(self, sequence_inputs, **kwargs):
        return self._encoding_func(sequence_inputs)

    def decode(self, response, **kwargs):
        return self._decoding_func(response)

    @abc.abstractmethod
    def _get_input_chats(self, inputs, conversation_start):
        """

        Get the raw input chats (and other inputs if any) from the inputs dict
        return the selected inputs split as sequence_inputs and other_inputs.

        sequence_inputs will be further processed, other_inputs not

        :param inputs: Dict with one or more different types of inputs
        :param conversation_start: Bool, if the given inputs are the initial inputs of a new
                                   conversation

        :return: equence_inputs, other_inputs
        """
        raise NotImplementedError("Please implement this method in a child class")

    @abc.abstractmethod
    def _build_input_sequences(self, sequence_inputs, conversation_start):
        raise NotImplementedError("Please implement this method in a child class")

    def _postprocess(self, sequence_inputs):
        """
        Optional step

        :param sequence_inputs:
        :return:
        """

        return sequence_inputs






from .simple_bot import SimpleBot

from .conversation_engine import ConversationEngineBase, BasicConversationEngine, UnableToGenerateValidResponse

from .io_processor import IOProcessorBase, SimpleIOProcessor

from .model_evaluator import ModelEvaluatorBase

from .sample_and_rank import SampleAndRank



import sys

import numpy as np


def response_score(scores):
    response_length = len(scores)
    if response_length == 0:
        return float('inf')

    response_prob = np.prod(scores)
    if response_prob == 0.0:
        return float('inf')
    else:
        return -np.sum(np.log(scores))/response_length


class SampleAndRank:

    def __init__(self, *args, num_candidate_responses=20, **kwargs):
        """
        Mixin for conversation engines to generate multiple candidate responses and choose the best one.

        :param num_candidate_responses:

        :param kwargs:
        """
        super().__init__(*args, **kwargs)

        self._num_candidate_responses = num_candidate_responses

    def respond(self, inputs, conversation_start=False):
        processed_inputs = self._io_processor.process_inputs(inputs, conversation_start)

        self._model_evaluator.update_context(processed_inputs, self._conversation_context, conversation_start)

        responses = []
        response_scores = []

        while len(responses) < self._num_candidate_responses:
            response, scores = self._create_response()

            if response in responses:
                continue

            responses += [response]
            response_scores += [response_score(scores)]

        sort_idx = np.argsort(response_scores)
        response_scores = [response_scores[i] for i in sort_idx]
        responses = [responses[i] for i in sort_idx]

        if self._debug:
            self._log.debug("Potential responses : ")
            for r, s in zip(responses, response_scores):
                s = round(s*100)/100 if s != float('inf') else "INF"
                sys.stdout.write(f"[{s}]\t: {r}\n")

        response_idx = np.argmin(response_scores)

        return responses[response_idx], [response_scores[response_idx]]




import torch
import torch.nn.functional as F


def random_sample(logits):
    """

    :param logits: shape (num_tokens,)
    :return: token probabilities, token index

    """
    p = F.softmax(logits, dim=0)

    token_idx = torch.multinomial(p.view(-1), 1)
    token_p = torch.gather(p, 0, token_idx)

    return token_p.squeeze(), token_idx.squeeze()


def select_max(logits):
    p = F.softmax(logits, dim=0)

    token_p, token_idx = torch.max(p, dim=0)

    return token_p, token_idx


def filter_top_p(logits, top_p=0.9):
    """

    :param logits: shape (num_tokens,)
    :param top_p:

    :return: kept logits, kept token indices
    """

    p = F.softmax(logits, dim=0)

    p_sorted, sorted_indices = torch.sort(p, dim=0, descending=True)
    p_cumulative = torch.cumsum(p_sorted, dim=0)

    to_keep = p_cumulative <= top_p

    # Always keep at least one, so shift to the right
    to_keep[1:] = to_keep[:-1].clone()
    to_keep[0] = True

    to_keep = sorted_indices[to_keep]

    return logits[to_keep], to_keep


def filter_top_k(logits, top_k=20):
    """

    :param logits: shape (num_tokens,)
    :param top_k:

    :return: kept logits, kept token indices
    """

    p = F.softmax(logits, dim=0)

    p_sorted, sorted_indices = torch.sort(p, dim=0, descending=True)

    to_keep = sorted_indices[:top_k]

    return logits[to_keep], to_keep



from botshop.conversation_engine import BasicConversationEngine as BasicConversationEngineBase


class BasicConversationEngine(BasicConversationEngineBase):

    def _unwrap(self, tensor):
        """
        Unwrap scalar tensor

        :param tensor:
        :return:
        """
        return tensor.item()



import abc
import sys

import re
import statistics

from .conversation_engine import UnableToGenerateValidResponse

from basics.base import Base

from basics.logging import get_logger
from basics.logging_utils import log_exception


def chat_with(bot, user_name="You", logger=None):
    if logger is None:
        logger = get_logger("Bot")

    while True:
        logger.info(f'{user_name} :')
        # 3) ask for input
        user_input = input()
        sys.stdout.write('\n')
        sys.stdout.flush()

        response = bot.respond_to(user_input)

        if response["system"] == "quit":
            break


class BotBase(Base, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def reset_state(self):
        raise NotImplementedError("Please implement this method in your child class")


    @abc.abstractmethod
    def respond_to(self, user_chat, user_name=None):
        raise NotImplementedError("Please implement this method in your child class")


class SimpleBot(BotBase):
    SYSTEM_COMMAND = re.compile('^##(.+)##$')

    RESET_STATE_COMMAND = re.compile('^(reset|r)$', re.I)

    QUIT_COMMAND = re.compile('^(q|quit)$', re.I)

    def __init__(self,
                 conversation_engine,
                 user_name='User',
                 bot_name='Bot',
                 init_chats=None,
                 init_is_user=None,
                 init_actor_name=None,
                 log_conversation=True,
                 debug=False,
                 name=None):
        """

        :param conversation_engine: Conversation engine instance derived from ConversationEngineBase
        :param user_name: Optional, default user name if no user name given with respond_to() method
        :param bot_name: Optional, default bot name
        :param init_chats: Optional, list, initial chat history
                           [
                                <chat 0>
                                ...
                                <chat N>
                           ]
        :param init_is_user: Optional, list, when init_chats is given, per chat it can be provided if it was a
                             user (True) or bot (False) utterance. len(init_is_user) must be equal to len(init_chats)
                             If not provided, it is assumed that the user started, and that the user and the bot take
                             equal turns after each other.
                             [
                                <True|False>
                                ...
                                <True|False>
                            ]
        :param init_actor_name: Optional, list, when init_chats is given, per chat the actor name, who uttered the chat,
                                can be provided. If this list is not provided, it is assumed that the user started and
                                that the bot and user take equal turns after each other; in this way the default
                                user name and bot name are used as actor names.
                                [
                                    <actor name 0>
                                    ...
                                    <actor name N>
                                ]

        :param log_conversation: If True, the conversations will be logged to stdout
        :param debug: If True, more logging is done
        :param name: Name of Bot system (class name by default)
        """

        super().__init__(pybase_logger_name=name)

        self._conversation_engine = conversation_engine

        self._user_name = user_name
        self._bot_name = bot_name

        self._log_conversation = log_conversation
        self._debug = debug

        # chat history
        if init_chats is None:
            init_chats = []
            init_is_user = []
            init_actor_name =[]
        else:
            if init_is_user is None:
                init_is_user = [bool(i % 2) for i in range(len(init_chats))]

            if init_actor_name is None:
                init_actor_name = [user_name if (i % 2) == 0 else bot_name for i in range(len(init_chats))]

            if len(init_chats) != len(init_is_user) != len(init_actor_name):
                raise Exception("Lists init_chats, init_is_user and init_actor_name nust all have the same length")

            if self._debug:
                self._log.debug("Initial chat history provided:")
                for actor_name, is_user, chat in zip(init_actor_name, init_is_user, init_chats):
                    self._log.debug(f"[{'USER' if is_user else 'BOT '}] {actor_name}: {chat}")

        self._init_chats = init_chats.copy()
        self._init_is_user = init_is_user.copy()
        self._init_actor_name = init_actor_name.copy()

        self._chats = None
        self._is_user = None
        self._actor_name = None

        self._conversation_start = True

        self.reset_state()

    def reset_state(self):
        if self._debug:
            self._log.debug("Resetting state ...")

        self._conversation_engine.reset_state()

        self._chats = self._init_chats.copy()
        self._is_user = self._init_is_user.copy()
        self._actor_name = self._init_actor_name.copy()

        self._conversation_state = None
        self._conversation_start = True

    def respond_to(self, user_chat, user_name=None):
        if user_name is None:
            user_name = self._user_name

        if self._log_conversation:
            self._log_user_input(user_chat, user_name)

        scores = None
        system_message = self._execute_command_in(user_chat, user_name=user_name)
        bot_chat = None
        other_outputs = None
        if system_message is None:  # No command executed
            self._chats.append(user_chat)
            self._is_user.append(True)
            self._actor_name.append(user_name)

            try:
                bot_chat, scores, other_outputs = self._respond()
            except UnableToGenerateValidResponse as e:
                system_message = f"Bot {self._bot_name} was unable to generate a valid response. " \
                                 f"Your chat is not recorded."
            except Exception as e:
                system_message = f"An unexpected exception occurred, bot {self._bot_name} was " \
                                 f"unable to generate a response. Your chat is not recorded."
                log_exception(self._log, "system_message", e)

            if system_message is None:  # No exception occurred
                if self._conversation_start:
                    self._conversation_start = False

                self._chats.append(bot_chat)
                self._is_user.append(False)
                self._actor_name.append(self._bot_name)

                if self._log_conversation:
                    self._log_response(self._bot_name, bot_chat, score=self._calc_final_response_score(scores))

        if self._log_conversation and system_message is not None:
            self._log_response("System", system_message)

        return {
            'bot': bot_chat,
            'scores': scores,
            'other_outputs': other_outputs,
            'system': system_message,
        }

    def _execute_command_in(self, user_chat, user_name=None):
        if user_name is None:
            user_name = self._user_name

        m = self.SYSTEM_COMMAND.match(user_chat)

        if m is None:
            return None

        try:
            command = m.groups()[0]

            m = self.QUIT_COMMAND.match(command)
            if m is not None:
                system_msg = "quit"
                return system_msg

            m = self.RESET_STATE_COMMAND.match(command)
            if m is not None:
                self.reset_state()
                system_msg = "Conversation model state has been reset."
                return system_msg

            return self._forward_command_to_conversation_engine(command, user_name=user_name)
        except Exception as e:
            log_exception(self._log, 'An exception occurred parsing the system command', e)
            return 'Unable to parse system command, no effect.'

    def _forward_command_to_conversation_engine(self, command, user_name=None):
        if user_name is None:
            user_name = self._user_name

        try:
            return self._conversation_engine.execute_command(command, user_name=user_name)
        except Exception as e:
            log_exception(self._log, 'Forwarding system command to evaluator failed', e)
            return 'Execution of system command by evaluator failed, no effect.', None

    def _respond(self):
        response, scores, other_outputs = self._conversation_engine.respond({
            "chats": self._chats,
            "is_user": self._is_user,
            "actor_name": self._actor_name,
            "responding_actor": self._bot_name
        }, self._conversation_start)

        return response, scores, other_outputs

    def _calc_final_response_score(self, scores):
        if scores is None:
            return None

        return statistics.mean(scores)

    def _log_user_input(self, user_chat, user_name=None):
        if user_name is None:
            user_name = "User"

        self._log.info(f'{user_name} : {user_chat}')

    def _log_response(self, actor_name, response, score=None):
        if score is not None:
            self._log.info('%s [%f] : \n%s\n' % (actor_name, score, response))
        else:
            self._log.info('%s : \n%s\n' % (actor_name, response))

